
#include "..\..\common\ImageFilter.h"


using namespace FilterSpace;
using namespace System::Windows::Forms;

IntPtr DumpErrorMessage(System::String^ err_msg)
{
	MessageBox::Show(err_msg, "Error", MessageBoxButtons::OK, MessageBoxIcon::Error);
	return IntPtr(0);
}


bool FilterCore(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if (!f)           
	{
		MessageBox::Show("FilterCore: !f");
		return false;
	}
	else if (!(f->datain)) 
	{
		MessageBox::Show("FilterCore: !f->datain");
		return false;
	}

	// BEGIN --> COMMON SUPPORT CODE (DO NOT MODIFY!)
	sw = f->width;
	sh = f->height;

	int   x, y;
	byte *din, *doutR, *doutG, *doutB;

	for (y = 0; y < sh; ++y)
	{
		din   = f->datain   + y * f->stridein;
		doutR = f->dataoutR + y * f->strideout;
		doutG = f->dataoutG + y * f->strideout;
		doutB = f->dataoutB + y * f->strideout;

		for (x = 0; x < sw; ++x)
		{
			*(doutB++) = (byte)(*din++);
			*(doutG++) = (byte)(*din++);
			*(doutR++) = (byte)(*din++);
		}

		// FOR PERFORMANCE MEASUREMENT
		if (f->processing_ptr) *(f->processing_ptr) += sw;
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}
	return true;
}


//
// *********
IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int indata = 0, outdataR = 0, outdataG = 0, outdataB = 0, width = 0, height = 0, strideout = 0;

	if (!fs) return IntPtr(0);
	xml_node ^xt, ^xt2, ^xt3, ^xn  = gcnew xml_node( fs );
	String   ^in_pin_name = "", ^out_pin_name = "";

	if (! xn->XmlGetChildByName(xt, "InputImage0" ))     return DumpErrorMessage("InputImage0 could not find.");
	if (! xt->XmlGetChildValueByName("Data",   indata) ) return DumpErrorMessage("Data could not find in InputImage0.");
	if (! xt->XmlGetChildValueByName("Width",  width ) ) return DumpErrorMessage("Width could not find in InputImage0.");
	if (! xt->XmlGetChildValueByName("Height", height) ) return DumpErrorMessage("Height could not find in InputImage0.");
	
	xt->XmlGetChildValueByName("PinName", in_pin_name);

	filtersettings *f = XmlToMyStruct(xn, true);

	if (width * height)
	{
		strideout = ((width + 3) >> 2) << 2;

		// Red
		byte *red = (xn->XmlGetChildByName(xt2, "OutputImage0")) ? f->dataoutR : 0;
		xn->XmlRemoveChildByName("OutputImage0");
		if (red == 0) red = new byte[strideout * height];
		if (in_pin_name != "") out_pin_name = in_pin_name + " (Red)";
		xn->children.Add(xt2 = gcnew xml_node(xt->XML_Tree_Marshal()));
		xt2->name = "OutputImage0";
		CreateImageSetting(xt2, out_pin_name, width, height, strideout, 1, 1, 1, 8, red );
		xt2->XmlGetChildByName( xt3, "ChannelsBitAllocation" );
		SetChannelsBitAllocation(xt3, 0, 8);

		// Green
		byte *green = (xn->XmlGetChildByName(xt2, "OutputImage1")) ? f->dataoutG : 0;
		xn->XmlRemoveChildByName("OutputImage1");
		if (green == 0) green = new byte[strideout * height];
		if (in_pin_name != "") out_pin_name = in_pin_name + " (Green)";
		xn->children.Add(xt2 = gcnew xml_node(xt->XML_Tree_Marshal()));
		xt2->name = "OutputImage1";
		CreateImageSetting(xt2, out_pin_name, width, height, strideout, 1, 1, 1, 8, green );
		xt2->XmlGetChildByName( xt3, "ChannelsBitAllocation" );
		SetChannelsBitAllocation(xt3, 0, 8);

		// Blue
		byte *blue = (xn->XmlGetChildByName(xt2, "OutputImage2")) ? f->dataoutB : 0;
		xn->XmlRemoveChildByName("OutputImage2");
		if (blue == 0) blue = new byte[strideout * height];
		if (in_pin_name != "") out_pin_name = in_pin_name + " (Blue)";
		xn->children.Add(xt2 = gcnew xml_node(xt->XML_Tree_Marshal()));
		xt2->name = "OutputImage2";
		CreateImageSetting(xt2, out_pin_name, width, height, strideout, 1, 1, 1, 8, blue );
		xt2->XmlGetChildByName( xt3, "ChannelsBitAllocation" );
		SetChannelsBitAllocation(xt3, 0, 8);

//		MessageBox::Show("CreateFilterSettings: silent=" + silent.ToString() + "\nWidth: " + width.ToString() + "\nHeight: " + height.ToString() + "\nData: " + indata.ToString() + "\nStride: " + strideout.ToString());
	}

	delete f;
	return (IntPtr)(void*)xn->XML_Tree_Marshal();
}



filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	int iTmp, packed, interleaved;
	xml_node^ xntmp;

	// Input

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;
	memset(filset, 0, sizeof(filtersettings));
	
	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain)) 
	{
		MessageBox::Show("XmlToMyStruct - 0 - " + creating.ToString());
		delete filset;
		return 0;
	}

	// Red
	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}
	else
	{
		if (GetOneImageSetting(xntmp, iTmp, iTmp, filset->strideout, iTmp, iTmp, iTmp, iTmp, filset->dataoutR))
		{
			MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
			delete filset;
			return 0;
		}
	}

	// Green
	if (! xn->XmlGetChildByName(xntmp, "OutputImage1") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 3 - " + creating.ToString());
		delete filset;
		return 0;
	}
	else
	{
		if (GetOneImageSetting(xntmp, iTmp, iTmp, iTmp, iTmp, iTmp, iTmp, iTmp, filset->dataoutG))
		{
			MessageBox::Show("XmlToMyStruct - 4 - " + creating.ToString());
			delete filset;
			return 0;
		}
	}

	// Blue
	if (! xn->XmlGetChildByName(xntmp, "OutputImage2") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 5 - " + creating.ToString());
		delete filset;
		return 0;
	}
	else
	{
		if (GetOneImageSetting(xntmp, iTmp, iTmp, iTmp, iTmp, iTmp, iTmp, iTmp, filset->dataoutB))
		{
			MessageBox::Show("XmlToMyStruct - 6 - " + creating.ToString());
			delete filset;
			return 0;
		}
	}

	filset->nCoreKillSignal = ( xn->XmlGetChildValueByName("nCore!KillSignal", iTmp) ) ? (int*)(void*)iTmp : 0;
	filset->processing_size = ( xn->XmlGetChildValueByName("Processing_Size",  iTmp) ) ? (int*)(void*)iTmp : 0;
	filset->processing_ptr  = ( xn->XmlGetChildValueByName("Processing_Ptr",   iTmp) ) ? (int*)(void*)iTmp : 0;

//	MessageBox::Show("XmlToMyStruct: Done");

	return filset;
}



bool ImageFilter::CanLocalize(){ return false; }

IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "1" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "3" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            3,        24,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);
	SetChannelsBitAllocation(xn, 1, 8);
	SetChannelsBitAllocation(xn, 2, 8);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName,       width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Red Channel",     0,      0,      0,      1,           1,            1,         8,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	xn = gcnew xml_node( node, "OutputImage1", "", "" );
	node->children.Add( xn );
	//                     PinName,         width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Green Channel",     0,      0,      0,      1,           1,            1,         8,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	xn = gcnew xml_node( node, "OutputImage2", "", "" );
	node->children.Add( xn );
	//                     PinName,        width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Blue Channel",     0,      0,      0,      1,           1,            1,         8,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);
	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}




// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	return (::FilterCore(f, sx, sy, sw, sh, nullptr));
}

bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;
	if (filset)
	{
		*(filset->processing_ptr) = 0;
		*(filset->processing_size) = filset->width * filset->height;

		b = ::FilterCore(filset, 0, 0, 0, 0, nullptr); // FASTCALL WITH 0,0,0,0 PARAMETERS

		*(filset->processing_ptr) = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
		delete filset;
	}
	return b;
}


